﻿/*
 * GPS3 Click driver for Mikrobus.Net
 * 
 * Version 1.0 :
 *  - Initial version coded by Stephen Cardinale.
 *  - Minimal GPS support for GPRMC sentences.
 * 
 * References needed :
 *  Microsoft.SPOT.Hardware
 *  Microsoft.SPOT.Hardware.SerialPort
 *  Microsoft.SPOT.Native
 *  MikroBusNet
 *  mscorlib
 *  
 * Source for SimpleSerial taken from IggMoe's SimpleSerial Class https://www.ghielectronics.com/community/codeshare/entry/644
 *  
 * Copyright 2014 MikroBus.Net
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
 * either express or implied. See the License for the specific language governing permissions and limitations under the License.
 * press or implied. See the License for the specific language governing permissions and limitations under the License.
 * 
 */

#region Using

using System;
using System.Collections;
using System.IO.Ports;
using System.Reflection;
using System.Threading;
using MBN.Enums;
using MBN.Exceptions;
using Microsoft.SPOT.Hardware;
using Math = System.Math;

#endregion

namespace MBN.Modules
{
    /// <summary>
    ///     Main class for the GPS3Click driver.
    ///     <para><b>Pins used :</b> Rx, Tx, Rst</para>
    ///     <para>The GPS3 click is a UART module and does not use Hardware Flow Control.</para>
    /// </summary>
    /// <remarks>
    /// <para>
    /// This driver supports only minimal features of the GPS3 Click.
    /// </para>
    /// <para>
    /// VCC should be connected to the VBAT pads to maintain the Volatile Memory (VM) in the GPS module.
    /// If it is not connected, the module will always perform a Full Cold Start once it is powered on because the navigation data and time stored in the VM will be lost,
    /// and this will result in longer Time To First Fix (TTFF). In addition, any changes to the parameter settings through command, such as changing baud rate are also stored in the VM,
    /// these configuration settings will be lost and be reverted back to the default firmware value if VBACKUP is not connected
    /// If VBACKUP is connected and the previous data stored in the VM are valid, the TTFF time will be drastically reduced (hot-start or warm-start, depending on the length of time between each power-on).
    /// Things like RTC almanac/ephemeris and last known position as well as some other proprietary data are stored inside the VM.
    /// Typically you should see the GPS module perform a hot start if it is powered on within 2 hours of being turned off. Assuming VBACKUP is connected, of course.
    /// The GPS module will perform warm-start if it is powered on after more than 2 hours.
    /// If the ephemeris/almanac is determined to be unusable, such as when the receiver is moved across great distances or the ephemeris/almanac simply expires, then the receiver will perform Cold Start.
    /// Typically you will see the module perform cold start if it has been off for several days.
    /// </para>
    /// </remarks>
    /// <example>Example Usage:
    /// <code language = "C#">
    /// using MBN;
    /// using System.Threading;
    /// using MBN.Exceptions;
    /// using MBN.Modules;
    /// using MBN.Modules.GHI;
    /// using Microsoft.SPOT;c
    /// namespace Examples
    /// {
    ///    public class Program
    ///    {
    ///        private static GPS3Click gps;
    ///        public static void Main()
    ///        {
    ///            try
    ///            {
    ///                gps = new GPS3Click(Hardware.SocketTwo);
    ///            }
    ///            catch (DeviceInitialisationException ex)
    ///            {
    ///                Debug.Print(ex.Message);
    ///                throw new DeviceInitialisationException("Unable to initialize GPS3 Click");
    ///            }
    ///
    ///            gps.PositionChange += GpsPositionChange;
    ///            gps.SerialError += gps_SerialError;
    ///
    ///            gps.StartGPS();
    ///
    ///            Thread.Sleep(Timeout.Infinite);
    ///        }
    ///
    ///        static void GpsPositionChange(GPS3Click sender, GPS3Click.Position position)
    ///        {
    ///            if (!position.DataValid) return;
    ///            Debug.Print("Position Changed");
    ///            Debug.Print(position.ToString());
    ///        }
    ///
    ///        static void gps_SerialError(object sender, string errorDescription, System.DateTime eventTime)
    ///        {
    ///            Debug.Print("Serial error - " + errorDescription);
    ///            gps.Serial.DiscardInBuffer();
    ///            gps.Serial.Flush();
    ///        }
    ///
    ///    }
    /// }
    /// </code>
    /// <code language = "VB">
    /// Option Explicit On
    /// Option Strict On
    /// Imports Microsoft.SPOT
    /// Imports MBN
    /// Imports System.Threading
    /// Imports MBN.Exceptions
    /// Imports MBN.Modules
    /// Namespace Examples
    ///     Public Module Module1
    ///         Dim WithEvents gps As GPS3Click
    ///         Sub Main()
    ///             Try
    ///                 gps = New GPS3Click(Hardware.SocketTwo)
    ///             Catch ex As DeviceInitialisationException
    ///                 Debug.Print(ex.Message)
    ///                 Throw New DeviceInitialisationException("Unable to initialize GPS3 Click")
    ///             End Try
    ///             gps.StartGPS()
    ///             Thread.Sleep(Timeout.Infinite)
    ///         End Sub
    ///         Private Sub gps_PositionChange(sender As GPS3Click, positionData As GPS3Click.Position) Handles gps.PositionChange
    ///             If (Not positionData.DataValid) Then Exit Sub
    ///             Debug.Print("Position Changed")
    ///             Debug.Print(positionData.ToString())
    ///         End Sub
    ///         Private Sub gps_SerialError(sender As Object, errorDescription As String, eventTime As Date) Handles gps.SerialError
    ///             Debug.Print("Serial error - " <![CDATA[&]]> errorDescription)
    ///             gps.Serial.DiscardInBuffer()
    ///             gps.Serial.Flush()
    ///         End Sub
    ///     End Module
    /// End Namespace
    /// </code>
    /// </example>
    public partial class GPS3Click : IDriver
    {
        #region Constants

        private const String EOL = "\r\n";
        private const String CMD_TEST_PACKET = "$PMTK000*32";
        private const String CMD_NMEA_RMC = "$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0";

        #endregion

        #region Fields

        private static SimpleSerial _serial;

        private readonly OutputPort _resetPin;

        private static readonly AutoResetEvent _autoEvent = new AutoResetEvent(false);
        private const UInt16 _autoResetWaitTime = 500;
        private static Boolean _deviceReady;

        private Antenna _antennaStatus = Antenna.Open;

        private Boolean processQueueMessages;
        private static readonly Queue messageQueue = new Queue();
      
        private TimeSpan lastPositionFixTime;

        private static Position _lastPosition = new Position();
        
        #endregion

        #region ENUMS

        /// <summary>
        ///     The current state of the External AntennaStatus circuitry.
        /// </summary>
        public enum Antenna
        {
            /// <summary>
            ///     External active antenna is connected and the module will use external active antenna.
            /// </summary>
            Ok,

            /// <summary>
            ///     External antenna is in open-circuit state (not connected) and the internal antenna is used at this time.
            /// </summary>
            Open,

            /// <summary>
            ///     External AntennaStatus is detected and short circuited. The internal antenna is being used.
            /// </summary>
            Shorted
        }

        /// <summary>
        ///     Position Fix acquisition type.
        /// </summary>
        public enum PositioningModes
        {
            /// <summary>
            ///     No Fix
            /// </summary>
            NoFix,

            /// <summary>
            ///     AutonomousFix GNSS fix
            /// </summary>
            AutonomousFix,

            /// <summary>
            ///     DifferentialFix GNSS fix
            /// </summary>
            DifferentialFix
        }

        /// <summary>
        ///     Fix Mode Manual (forced 3D) or Auto 2D/3D.
        /// </summary>
        public enum FixMode
        {
            /// <summary>
            ///     Manual, forced to switch to 3D mode
            /// </summary>
            Manual,

            /// <summary>
            ///     Allowed to automatically switch between 2D/3D modes
            /// </summary>
            Auto,
        }

        /// <summary>
        ///     GPS Fix Status
        /// </summary>
        public enum FixStatus
        {
            /// <summary>
            ///     No fix
            /// </summary>
            NoFix,

            /// <summary>
            ///     2D fix
            /// </summary>
            Fix2D,

            /// <summary>
            ///     3D fix
            /// </summary>
            Fix3D
        }

        #endregion

        #region CTOR

        /// <summary>
        ///     Initializes a new instance of the <see cref="GPS3Click" /> class.
        /// </summary>
        /// <param name="socket">The socket on which the GPS3Click module is plugged on MikroBus.Net board</param>
        public GPS3Click(Hardware.Socket socket) 
        {
            try
            {
                Hardware.CheckPins(socket, socket.Rx, socket.Tx, socket.Rst);

                _resetPin = new OutputPort(socket.Rst, true); // Reset circuit is active low
               
                _serial = new SimpleSerial(socket.ComPort, 9600) { ReadTimeout = 0 };
                
                if (!_serial.IsOpen) _serial.Open();
                _serial.Flush();
                _serial.DiscardInBuffer();
                _serial.DiscardOutBuffer();
                _serial.DataReceived += _serial_DataReceived;
                _serial.ErrorReceived += _serial_ErrorReceived;

                InternalReset();

                _serial.Write("$PMTK000*32" + EOL + EOL); // Can't use SendCommand here as the device has not been initialized yet. Note - If you are going to use the Write method, you need to send two EOLs.
               
                // Check if device is connected and communicating. 
                _autoEvent.Reset();
                _autoEvent.WaitOne(_autoResetWaitTime, true);
                if (!_deviceReady) throw new DeviceInitialisationException("Error initializing the GPS3");

                // Subscribe only to GPRMC sentences.
                SendCommand(CMD_NMEA_RMC);
                
                lastPositionFixTime = TimeSpan.MinValue;

                new Thread(ProcessMessageQueue).Start();
            }
            //Catch only the PinInUse exception, so that program will halt on other exceptions send it directly to the caller
            catch (PinInUseException ex)
            {
                throw new PinInUseException(ex.Message);
            }
        }
        
        #endregion

        #region Properties

        /// <summary>
        /// Gets the last RMC Data received from the GPS3 Click.
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// Debug.Print("Last RMC Data");
        /// Debug.Print(gps.LastPosition.ToString());
        /// </code>
        /// <code language = "VB">
        /// Debug.Print("Last RMC Data")
        /// Debug.Print(gps.LastPosition.ToString())
        /// </code>
        /// </example>
        public Position LastPosition
        {
            get { return _lastPosition; }
        }

        /// <summary>
        ///     Gets the elapsed time since the last RMC position was received.
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// Debug.Print("Tim since last valid RMC fix - " + gps.LastValidFixAge);
        /// </code>
        /// <code language = "VB">
        /// Debug.Print("Tim since last valid RMC fix - " <![CDATA[&]]> gps.LastValidFixAge)
        /// </code>
        /// </example>
        public TimeSpan LastValidFixAge
        {
            get
            {
                if (lastPositionFixTime == TimeSpan.MinValue) return TimeSpan.MaxValue;
                return Utility.GetMachineTime() - lastPositionFixTime;
            }
        }

        /// <summary>
        ///     Gets the underlying Serial Port Instance of the driver of which you can manipulate based upon your needs.
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// // Put the GPS3 Click in Standby Mode.
        /// var commandString = "$PMTK161,0*";
        /// commandString += gps.CalculateChecksum(commandString + "\r\n" + "\r\n");
        /// var commandBytes = Encoding.UTF8.GetBytes(commandString);
        /// gps.Serial.Write(commandBytes, 0, commandBytes.Length);
        /// </code>
        /// <code language = "VB">
        /// ' Put the GPS3 Click in Standby Mode.
        /// Dim commandString As String = "$PMTK161,0*"
        /// commandString += gps.CalculateChecksum(commandString <![CDATA[&]]> Microsoft.VisualBasic.Constants.vbCrLf <![CDATA[&]]> Microsoft.VisualBasic.Constants.vbCrLf)
        /// Dim commandBytes As Byte() = Encoding.UTF8.GetBytes(commandString)
        /// gps.Serial.Write(commandBytes, 0, commandBytes.Length)
        /// </code>
        /// </example>
        public SerialPort Serial
        {
            get { return _serial; }
        }

        /// <summary>
        ///     Gets the <see cref="MBN.Modules.GPS3Click.Antenna" /> of the External Antenna connection.
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// var status = gps.AntennaStatus;
        /// Debug.Print("AntennaStatus is " + (status == GPS3Click.AntennaStatus.Ok ? "Open" : status == GPS3Click.AntennaStatus.Ok ? "Ok" : "Shorted"));
        /// </code>
        /// <code language = "VB">
        /// Dim status As GPS3Click.AntennaStatus = gps.AntennaStatus
        /// Debug.Print("AntennaStatus is " <![CDATA[&]]> If(status = GPS3Click.AntennaStatus.Ok, "Open", If(status = GPS3Click.AntennaStatus.Ok, "Ok", "Shorted")))
        /// </code>
        /// </example>
        public Antenna AntennaStatus
        {
            get { return _antennaStatus; }
        }
        
        /// <summary>
        ///     Gets or sets the power mode.
        /// </summary>
        /// <example>
        ///     None provided as this module does not support changing PowerModes directly. 
        /// </example>
        /// <exception cref="NotImplementedException">A NotImplementedException will be thrown if attempting to change the PowerMode of the GPS3 Click as the module doesn't support power modes.</exception>
        /// <example>Example Usage: None provided as this module does not support Power modes.
        /// </example>
        public PowerModes PowerMode
        {
            get { return PowerModes.On; }
            set
            {
                throw new NotImplementedException("Changing PowerMode is not supported.");
            }
        }

        /// <summary>
        ///     Gets the driver version.
        /// </summary>
        /// <example>Example usage:
        /// <code language="C#">
        /// Debug.Print ("Current driver version : "+ _gps.DriverVersion);
        /// </code>
        /// <code language = "VB">
        /// Debug.Print("Current driver version : " <![CDATA[&]]> _gps.DriverVersion.ToString())
        /// </code>
        /// </example>
        public Version DriverVersion
        {
            get { return Assembly.GetExecutingAssembly().GetName().Version; }
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Resets the module
        /// </summary>
        /// <param name="resetMode">The <see cref="ResetModes"/>to use to reset the module </param>
        /// <exception cref="NotImplementedException">A NotImplementedException will be thrown if attempting to REset the GPS3 Click as the module doesn't support power modes.</exception>
        /// <example>Example Usage: None provided as this module does not support Reset methods.
        /// </example>
        public Boolean Reset(ResetModes resetMode)
        {
            throw new NotSupportedException("Reset functionality is not supported by the GPS3 Click. To reset the GPS3 Click, use PerformHotStart, PreformWarmStart, PerformColdStart or PerformFullColdStart(FactoryReset) methods.");
        }

        /// <summary>
        ///     Sends a properly formatted NMEA Sentence (command) to the GPS3 Click appending the calculated checksum and the
        ///     EndOfLine marker "CRLF"
        /// </summary>
        /// <param name="command">
        ///     Then NMEA Command string to send, expected format is
        ///     "$PMTK314,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0"
        /// </param>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// //Send command to put GPS3 Click in Standby Low Power Mode.
        /// gps.SendCommand("$PMTK161,0");
        /// </code>
        /// <code language = "VB">
        /// 'Send command to put GPS3 Click in Standby Low Power Mode.
        /// gps.SendCommand("$PMTK161,0")
        /// </code>
        /// </example>
        public Boolean SendCommand(String command)
        {
            if (!_deviceReady) return false;
            _serial.WriteLine(command + "*" + CalculateChecksum(command) + EOL);
            return true;
        }

        /// <summary>
        ///     Validates NMEA string to verify if the checksum is correct, format of NEMA string is expected to be
        ///     "$PMTK001,869,3*37"
        /// </summary>
        /// <param name="nemaString">The NMEA string to verify.</param>
        /// <returns>Return true if the checksum is valid, or otherwise false.</returns>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// Debug.Print("ChecksumValid? " + gps.IsValidChecksum("$PMTK001,869,3*37"));
        /// </code>
        /// <code language = "VB">
        /// Debug.Print("ChecksumValid? " <![CDATA[&]]> gps.IsValidChecksum("$PMTK001,869,3*37"))
        /// </code>
        /// </example>
        public Boolean IsValidChecksum(String nemaString)
        {
            if (nemaString.IndexOf('*') <= -1) return false;
            nemaString = Replace(nemaString, " ", String.Empty);

            String valid_checksum = nemaString.Substring(nemaString.IndexOf('*') + 1, 2);
            nemaString = nemaString.Substring(1, nemaString.IndexOf('*') - 1);

            Int32 checksum = 0;

            for (Int32 i = 0; i < nemaString.Length; i++)
            {
                checksum ^= nemaString[i];
            }
            var returnValue = false;
            try
            {
                returnValue = checksum == Convert.ToInt32(valid_checksum, 16);
            }
            catch (ArgumentException)
            {
               // Sanity check
            }
            return returnValue;
        }

        /// <summary>
        ///     Calculates the XOR Checksum of a NEMA String, format of NEMA String is expected to be "$PMTK001,869,3*37"
        /// </summary>
        /// <param name="nemaString">The NMEA command string to calculate the checksum for.</param>
        /// <returns>Returns a <see cref="System.String" /> of the checksum.</returns>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// // Put the GPS3 Click in Standby Mode.
        /// var commandString = "$PMTK161,0*";
        /// commandString += gps.CalculateChecksum(commandString + "\r\n" + "\r\n");
        /// var commandBytes = Encoding.UTF8.GetBytes(commandString);
        /// gps.Serial.Write(commandBytes, 0, commandBytes.Length);
        /// </code>
        /// <code language = "VB">
        /// ' Put the GPS3 Click in Standby Mode.
        /// Dim commandString As String = "$PMTK161,0*"
        /// commandString += gps.CalculateChecksum(commandString + Microsoft.VisualBasic.Constants.vbCrLf + Microsoft.VisualBasic.Constants.vbCrLf)
        /// Dim commandBytes As Byte() = Encoding.UTF8.GetBytes(commandString)
        /// gps.Serial.Write(commandBytes, 0, commandBytes.Length)
        /// </code>
        /// </example>
        public String CalculateChecksum(String nemaString)
        {
            nemaString = Replace(nemaString, " ", String.Empty);
            nemaString = nemaString.Substring(nemaString.IndexOf('$') + 1);

            Int32 check = 0;
            for (Int32 c = 0; c < nemaString.Length; c++)
            {
                check ^= nemaString[c];
            }
            return check.ToString("X2");
        }

        /// <summary>
        /// Starts the GPS3 Click driver to process NMEA Messages.
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        ///  gps.StartGPS();
        /// </code>
        /// <code language = "VB">
        ///  gps.StartGPS()
        /// </code>
        /// </example>
        public void StartGPS()
        {
            processQueueMessages = true;
        }
        
        /// <summary>
        /// Stops the GPS3 Click driver to process NMEA Messages.
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        ///  gps.StopGPS();
        /// </code>
        /// <code language = "VB">
        ///  gps.StopGPS()
        /// </code>
        /// </example>
        public void StopGPS()
        {
            processQueueMessages = false;
        }

        #endregion

        #region Private Methods

        private void InternalReset()
        {
            _resetPin.Write(false);
            Thread.Sleep(15); // 10ms required per L80 Hardware Design v 1.2 Datasheet
            _resetPin.Write(true);
            Thread.Sleep(1000);
        }

        private static String Replace(String Source, String ToFind, String ReplaceWith)
        {
            Int32 startIndex = 0;
            if (Source == String.Empty || Source == null || (ToFind == String.Empty || ToFind == null)) return Source;
            while (true)
            {
                Int32 length = Source.IndexOf(ToFind, startIndex);
                if (length >= 0)
                {
                    Source = length <= 0
                        ? ReplaceWith + Source.Substring(length + ToFind.Length)
                        : Source.Substring(0, length) + ReplaceWith + Source.Substring(length + ToFind.Length);
                    startIndex = length + ReplaceWith.Length;
                }
                else
                {
                    break;
                }
            }
            return Source;
        }

        private static DateTime ParseDateTime(String utcTimeString, String utcDateString)
        {
            double timeRawDouble = Double.Parse(utcTimeString);
            var timeRaw = (Int32) timeRawDouble;
            int hours = timeRaw/10000;
            int minutes = (timeRaw/100)%100;
            int seconds = timeRaw%100;
            var milliseconds = (Int32) ((timeRawDouble - timeRaw)*1000.0);
            int dateRaw = Int32.Parse(utcDateString);
            int day = dateRaw/10000;
            int month = (dateRaw/100)%100;
            int year = 2000 + (dateRaw%100);
            return new DateTime(year, month, day, hours, minutes, seconds, milliseconds);
        }

        private static Double ParseLatitude(String latitudeString, String hemisphereString)
        {
            double latitudeRaw = Double.Parse(latitudeString);
            int latitudeDegreesRaw = ((Int32) latitudeRaw)/100;
            Double latitudeMinutesRaw = latitudeRaw - (latitudeDegreesRaw*100);
            double latitude = latitudeDegreesRaw + (latitudeMinutesRaw/60.0);
            if (hemisphereString == "S") latitude = -latitude;
            return latitude;
        }

        private static Double ParseLongitude(String longitudeString, String hemisphereString)
        {
            double longitudeRaw = Double.Parse(longitudeString);
            int longitudeDegreesRaw = ((Int32) longitudeRaw)/100;
            Double longitudeMinutesRaw = longitudeRaw - (longitudeDegreesRaw*100);
            double longitude = longitudeDegreesRaw + (longitudeMinutesRaw/60.0);
            if (hemisphereString == "W") longitude = -longitude;
            return longitude;
        }
        
        private void ProcessMessageQueue()
        {
            while (true)
            {
                if (messageQueue.Count == 0) continue;
                if (!processQueueMessages) continue;
                if (!_deviceReady) continue;

                ThreadStart threadStart;

                var sentence = (String)messageQueue.Dequeue();
                if (sentence == null || !IsValidChecksum(sentence)) continue;
                sentence = sentence.Substring(0, sentence.Length - 3);


                switch (sentence.Substring(0, 6))
                {
                    case "$GPRMC":
                    {
                        threadStart = () => ParseRMCSentence(sentence);
                        break;
                    }
                    case "$GPTXT":
                    {
                        threadStart = () => ParseTXTSentence(sentence);
                        break;
                    }
                    default:
                    {
                        continue;
                    }
                }

                if (threadStart.Method != null)
                {
                    var thread = new Thread(threadStart) { Priority = ThreadPriority.Normal };
                    thread.Start();
                }

                Thread.Sleep(20);
            }
        }
        
        private void _serial_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            var gpsDataIn = _serial.Deserialize();

            foreach (String gpsSentence in gpsDataIn)
            {
                if (gpsSentence == null || !IsValidChecksum(gpsSentence)) return;

                String sentence = gpsSentence.Substring(0, gpsSentence.Length - 3);

                if (sentence.Substring(0, 5) == "$PMTK")
                {
                    var pmtkThreadStart = new ThreadStart(() => ParsePMTKSentence(sentence));
                    var pmtkThread = new Thread(pmtkThreadStart) { Priority = ThreadPriority.Highest };
                    pmtkThread.Start();
                }

                messageQueue.Enqueue(gpsSentence);

                OnNMEASentenceReceived(this, sentence);
            }
        }

        private void ParseTXTSentence(String txtSentence)
        {
            string[] tokens = txtSentence.Split(',');
            if (tokens[4].Substring(0, 9) == "ANTSTATUS")
            {
                var s = tokens[4].Substring(10, tokens[4].Length - 10);
                var antennaStatus = s == "OK" ? Antenna.Ok : s == "OPEN" ? Antenna.Open : Antenna.Shorted;
                if (_antennaStatus == antennaStatus) return;
                _antennaStatus = antennaStatus;
                OnAntennaStatusChanged(this, antennaStatus);
            }
        }

        private void ParseRMCSentence(String rmcSentence)
        {
            string[] tokens = rmcSentence.Split(',');

            var latitude = ParseLatitude(tokens[3], tokens[4]);
            var longitude = ParseLongitude(tokens[5], tokens[6]);
         
            var rmcData = new Position
            {
                DateTimeUTC = ParseDateTime(tokens[1], tokens[9]),
                DataValid = tokens[2]== "A",
                COG = (tokens[8] != "") ? Double.Parse(tokens[8]) : 0,
                Latitude = (tokens[4] == "S" ? Math.Abs(latitude) : latitude),
                LatitudeHemisphere = tokens[4],
                Longitude = (tokens[6] == "W") ? Math.Abs(longitude) : longitude,
                LongitudeHemisphere = tokens[6],
                Speed = (tokens[7] != "") ? Double.Parse(tokens[7]) : 0,
                PositioningMode = tokens[12] == "D" ? PositioningModes.DifferentialFix : tokens[12] == "A" ? PositioningModes.AutonomousFix : PositioningModes.NoFix
            };

            if ((Math.Abs(_lastPosition.Latitude - rmcData.Latitude) > 0.0) || Math.Abs(_lastPosition.Longitude - rmcData.Longitude) > 0.0) OnCoordinatesUpdated(this, rmcData.Latitude, rmcData.Longitude);
            if (Math.Abs(_lastPosition.Speed - rmcData.Speed) > 0.0) OnSpeedChanged(this, rmcData.Speed);
            if (Math.Abs(_lastPosition.COG - rmcData.COG) > 0.0) OnCourseChanged(this, rmcData.COG);
            if (_lastPosition.PositioningMode != rmcData.PositioningMode) OnRMCPositioningModeChanged(this, rmcData.PositioningMode);

            if (_lastPosition.GetHashCode() != rmcData.GetHashCode())
            {
                _lastPosition = rmcData;
                OnPositionChanged(this, _lastPosition);
            }

            if (rmcData.DataValid) lastPositionFixTime = Utility.GetMachineTime();
        }

        private void ParsePMTKSentence(String pmtkSentence)
        {
            string[] tokens = pmtkSentence.Split(',');

            switch (tokens[0])
            {
                case "$PMTK001":
                {
                    switch (tokens[1])
                    {
                        case "0": // Packet Type: Test Packet (Undocumented)
                        {
                            // Response to Test packet ($PMTK000*32) - $PMTK001,0,3*30 means Device Ready and communicating at the selected BaudRate.
                            _deviceReady = tokens[2] == "3";
                            _autoEvent.Set();
                            break;
                        }
                    }
                    break;
                }
            }
            OnPMTKSentenceReceived(this, pmtkSentence);
        }

        #endregion

        #region Events

        #region AntennaStatus Status Changed Event

        /// <summary>
        ///     The delegate used for the <see cref="AntennaStatusChanged" /> event.
        /// </summary>
        /// <param name="sender">The object that raised the event.</param>
        /// <param name="status">The <see cref="Antenna" /> of the External AntennaStatus.</param>
        public delegate void AntennaStatusChangedHandler(GPS3Click sender, Antenna status);

        /// <summary>
        ///     The event that is raised when the status of the External AntennaStatus changes.
        /// </summary>
        public event AntennaStatusChangedHandler AntennaStatusChanged;

        private void OnAntennaStatusChanged(GPS3Click sender, Antenna status)
        {
            AntennaStatusChangedHandler handler = AntennaStatusChanged;
            if (handler == null) return;
            handler(sender, status);
        }

        #endregion

        #region MMEA Message Received Event

        /// <summary>
        ///     Represents the delegate that is used to handle the <see cref="GPS3Click.NMEASentenceReceived" /> event.
        /// </summary>
        /// <param name="sender">The object that raised the event.</param>
        /// <param name="nmeaSentence">The event arguments.</param>
        public delegate void NMEASentenceReceivedHandler(GPS3Click sender, String nmeaSentence);

        /// <summary>
        ///     The event that is raised when an NMEA sentence is received.
        /// <para><b>This is for advanced users who want to parse the NMEA sentences themselves.</b></para>  
        /// </summary>
        public event NMEASentenceReceivedHandler NMEASentenceReceived;

        private void OnNMEASentenceReceived(GPS3Click sender, String nmeaSentence)
        {
            NMEASentenceReceivedHandler handler = NMEASentenceReceived;
            if (handler == null) return;
            handler(sender, nmeaSentence);
        }

        #endregion

        #region PMTK Message Received Event

        /// <summary>
        /// The delegate used for the <see cref="PMTKSentenceReceived"/> event.
        /// </summary>
        /// <param name="sender">The GPS3 click that raised the event.</param>
        /// <param name="pmtkSentence">The PMTK GPS Sentence</param>
        public delegate void PMTKSentenceReceivedHandler(GPS3Click sender, String pmtkSentence);

        /// <summary>
        ///     Raised when a PMTK sentence is received.  
        /// <para><b>This is for advanced users who want to parse the NMEA sentences themselves.</b></para>  
        /// </summary>
        public event PMTKSentenceReceivedHandler PMTKSentenceReceived;

        private void OnPMTKSentenceReceived(GPS3Click sender, String pmtkSentence)
        {
            PMTKSentenceReceivedHandler handler = PMTKSentenceReceived;
            if (handler == null) return;
            handler(sender, pmtkSentence);
        }

        #endregion

        #region Error Event

        /// <summary>
        ///    The delegate used for the <see cref="GPS3Click.SerialError" /> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event.</param>
        /// <param name="errorDescription">The <see cref="System.String" /> containing the description of the error."</param>
        /// <param name="eventTime">The time that the event occurred.</param>
        public delegate void ErrorReceivedHandler(object sender, String errorDescription, DateTime eventTime);

        /// <summary>
        ///     The event that is raised when the GPS3 Click receives an Error in data transmission.
        /// </summary>
        public event ErrorReceivedHandler SerialError;

        private void _serial_ErrorReceived(object sender, SerialErrorReceivedEventArgs e)
        {
            String eventType;
            switch (e.EventType)
            {
                case System.IO.Ports.SerialError.Frame:
                    eventType = "Frame Error";
                    break;
                case System.IO.Ports.SerialError.Overrun:
                    eventType = "Overrun Error";
                    break;
                case System.IO.Ports.SerialError.RXOver:
                    eventType = "RX Overflow Error";
                    break;
                case System.IO.Ports.SerialError.RXParity:
                    eventType = "RX Parity Error";
                    break;
                case System.IO.Ports.SerialError.TXFull:
                    eventType = "TX Buffer Full";
                    break;
                default:
                    eventType = e.ToString();
                    break;
            }
            ErrorReceivedHandler handler = SerialError;
            if (handler == null) return;
            handler(this, eventType, DateTime.Now);
        }

        #endregion

        #region Position Changed Event

        /// <summary>
        ///     The event that is raised when valid RMC Data is received.
        /// </summary>
        public event PositionChangedHandler PositionChange;

        /// <summary>
        ///     The delegate used for the <see cref="GPS3Click.PositionChange" /> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event.</param>
        /// <param name="positionData">The position data. See <see cref="Position"/></param>
        public delegate void PositionChangedHandler(GPS3Click sender, Position positionData);

        private void OnPositionChanged(GPS3Click sender, Position rmcData)
        {
            PositionChangedHandler handler = PositionChange;
            if (handler == null) return;
            handler(sender, rmcData);
        }

        #endregion

        #region Coordinates Changed Event

        /// <summary>
        /// The delegate used for the <see cref="GPS3Click.PositioningModeChanged"/> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event</param>
        /// <param name="latitude">The latitude of the GPS Fix</param>
        /// <param name="longitude">The Longitude of the GPS Fix.</param>
        public delegate void CoordinatesUpdatedHandler(GPS3Click sender, Double latitude, Double longitude);

        /// <summary>
        ///     The event that is raised when the Coordinates are changed.
        /// </summary>
        public event CoordinatesUpdatedHandler CoordinatesChanged;

        private void OnCoordinatesUpdated(GPS3Click sender, Double latitude, Double longitude)
        {
            CoordinatesUpdatedHandler handler = CoordinatesChanged;
            if (handler == null) return;
            handler(sender, latitude, longitude);
        }

        #endregion

        #region Course Changed Event

        /// <summary>
        /// The delegate used for the <see cref="GPS3Click.CourseChanged"/> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event</param>
        /// <param name="heading">The new heading in degrees.</param>
        public delegate void CourseChangedHandler(GPS3Click sender, Double heading);


        /// <summary>
        ///     The event that is raised when the Course (COG) has changed.
        /// </summary>
        public event CourseChangedHandler CourseChanged;

        private void OnCourseChanged(GPS3Click sender, Double newHeading)
        {
            CourseChangedHandler handler = CourseChanged;
            if (handler == null) return;
            handler(sender, newHeading);
        }

        #endregion

        #region Speed Changed Event

        /// <summary>
        /// The delegate used for the <see cref="GPS3Click.SpeedChanged"/> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event</param>
        /// <param name="knots">The speed in knots. One Knot or Nautical Mile per Hour is equivalent to 1.151 mph.</param>
        public delegate void SpeedChangedHandler(GPS3Click sender, Double knots);

        /// <summary>
        ///     The event that is raised when the Speed has changed.
        /// </summary>
        public event SpeedChangedHandler SpeedChanged;

        private void OnSpeedChanged(GPS3Click sender, Double knots)
        {
            SpeedChangedHandler handler = SpeedChanged;
            if (handler == null) return;
            handler(sender, knots);
        }

        #endregion  

        #region Fix Mode Change Event

        /// <summary>
        /// The delegate used for the <see cref="GPS3Click.CoordinatesChanged"/> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event</param>
        /// <param name="fixMode">The new <see cref="FixMode"/>.</param>
        public delegate void FixModeChangedHandler(GPS3Click sender, FixMode fixMode);
        
        /// <summary>
        ///     The event that is raised when the GNSS Fix Mode has changed.
        /// </summary>
        public event FixModeChangedHandler FixModeChanged;

        private void OnFixModeChanged(GPS3Click sender, FixMode fixMode)
        {
            FixModeChangedHandler handler = FixModeChanged;
            if (handler == null) return;
            handler(sender, fixMode);
        }

        #endregion

        #region Positioning Mode Changed Event

        /// <summary>
        /// The delegate used for the <see cref="GPS3Click.PositioningModeChanged"/> event.
        /// </summary>
        /// <param name="sender">The GPS3 Click that raised the event</param>
        /// <param name="positioningMode">The new <see cref="PositioningModes"/>.</param>
        public delegate void PositioningModeChangedHandler(GPS3Click sender, PositioningModes positioningMode); 
        /// <summary>
        /// The event that is raised when the <see cref="PositioningModes"/> has changed.
        /// </summary>
        /// <remarks>This event will only fire when subscribing to RMC Sentences.</remarks>
        public event PositioningModeChangedHandler PositioningModeChanged;

        private void OnRMCPositioningModeChanged(GPS3Click sender, PositioningModes positioningMode)
        {
            PositioningModeChangedHandler handler = PositioningModeChanged;
            if (handler == null) return;
            handler(sender, positioningMode);
        }

        #endregion
        
        #endregion

        #region Position Class

        /// <summary>
        ///     This class represents a GPS Position from a GPRMC sentence.
        ///     RMC, recommended minimum position data (including position, velocity, course and time).
        /// </summary>
        /// <example>Example Usage:
        /// <code language = "C#">
        /// </code>
        /// <code language = "VB">
        /// </code>
        /// </example>
        public class Position
        {
            /// <summary>
            ///     DateTime of the GPS Fix
            /// </summary>
            public DateTime DateTimeUTC { get; internal set; }

            /// <summary>
            ///     Valid data received. 
            /// </summary>
            public Boolean DataValid { get; internal set; }

            /// <summary>
            ///     The latitude.
            /// </summary>
            public Double Latitude { get; internal set; }

            /// <summary>
            ///     The North/South Hemisphere of the Latitude.
            /// </summary>
            public String LatitudeHemisphere { get; internal set; }

            /// <summary>
            ///     The longitude.
            /// </summary>
            public Double Longitude { get; internal set; }

            /// <summary>
            ///     The East/West Hemisphere of the Longitude.
            /// </summary>
            public String LongitudeHemisphere { get; internal set; }

            /// <summary>
            ///     SpeedKnotsRmc over the ground in knots.
            /// </summary>
            public Double Speed { get; internal set; }

            /// <summary>
            ///     CourseOverGround (COG) or heading in degrees.
            /// </summary>
            public Double COG { get; internal set; }

            /// <summary>
            ///     Position Fix acquisition type.
            /// </summary>
            public PositioningModes PositioningMode { get; internal set; }

            /// <summary>
            ///     Provides a formatted string for Position Data.
            /// </summary>
            /// <returns>The formatted string.</returns>
            public override String ToString()
            {
                return "DateTime: " + DateTimeUTC + ", Latitude: " + Latitude + " - " + LatitudeHemisphere + ", Longitude: " +
                       Longitude + " - " + LongitudeHemisphere + ", Speed(Knots): " + Speed + ", COG: " + COG +
                       ", Positioning Mode: " +
                       (PositioningMode == PositioningModes.AutonomousFix
                           ? "Autonomous"
                           : PositioningMode == PositioningModes.DifferentialFix ? "Differential" : "NoFix");
            }

            /// <summary>
            /// Serves as a hash function for a particular type. 
            /// </summary>
            /// <returns>
            /// A hash code for the current object.
            /// </returns>
            public override Int32 GetHashCode()
            {
                Double hash = 0;
                if (Longitude > 0) hash = Longitude;
                if (Latitude > 0) hash *= Latitude;
                if (Speed > 0) hash *= Speed;
                if (COG > 0) hash *= COG;
                if (LongitudeHemisphere == "W")
                {
                    hash *= 2;
                }
                else
                {
                    hash *= 1;
                }
                if (LatitudeHemisphere == "W")
                {
                    hash *= 2;
                }
                else
                {
                    hash *= 1;
                }
                return (Int32)hash;
            }

        }

        #endregion

    }
}